#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import logging
import operator
import re
from collections import defaultdict
from datetime import timezone, datetime
from typing import Optional, List, Dict, Tuple, Iterator, Union

from botocore.credentials import RefreshableCredentials
from botocore.session import get_session

from boto3 import Session
from dateutil.parser import isoparse

trace_logger = logging.getLogger('trace_logger')


class SplitString:
    """Field that serializes to a title case string and deserializes
    to a lower case string.
    """
    _re_match = re.compile(
        r'^([\sa-zA-z0-9_-]+)'
        '(<=|<|=|>|>=|!=)((?!<=|<|=|>|>=|!=)'
        r'[\sa-zA-z0-9_-]+)$'
    )

    def __init__(self, val):
        matches = SplitString._re_match.match(val)
        self.valid = True
        if not matches or len(matches.groups()) != 3:
            self.valid = False
            raise ValueError(
                f"Requested partition is invalid: {val}. Partition arguments "
                f"must be alphanumeric separated by an operator, and must "
                f"not be wrapped in special characters like single or double "
                f"quotations."
            )
        else:
            self.partition, self.operator, self.value = matches.groups()

    def as_dict(self):
        if self.valid:
            return {
                'partition': str(self.partition),
                'operator': str(self.operator),
                'value': str(self.value)
            }

        return None


def create_refreshing_session(
    dli_client_session, **kwargs
) -> Session:
    """
    :param dli_client_session: We must not be closing over the original
    variable in a multi-threaded env as the state can become shared.
    :param kwargs:
    :return:
    """

    def refresh():
        auth_key = dli_client_session.auth_key
        expiry_time = dli_client_session.token_expires_on.replace(
            tzinfo=timezone.utc
        ).isoformat()

        return dict(
            access_key=auth_key,
            secret_key='noop',
            token='noop',
            expiry_time=expiry_time,
        )

    _refresh_meta = refresh()

    session_credentials = RefreshableCredentials.create_from_metadata(
        metadata=_refresh_meta,
        refresh_using=refresh,
        method='noop'
    )

    session = get_session()
    handler = kwargs.pop("event_hooks", None)
    if handler is not None:
        session.register(f'before-send.s3', handler)

    session._credentials = session_credentials
    return Session(
        botocore_session=session,
        **kwargs
    )


def _get_partitions_in_filepath(filepath: str) -> List[List[str]]:
    splits = filepath.split("/")
    return [x.split("=") for x in splits if "=" in x]


def _operator_lookup(op):
    return {
        '<': operator.lt,
        '>': operator.gt,
        '=': operator.eq,
        '<=': operator.le,
        '>=': operator.ge,
        '!=': operator.ne
    }[op]


def eval_logical(oper, field, val):
    return _operator_lookup(oper)(field, val)


def match_partitions(
    file_path: str,
    partitions: Optional[List[str]],
):
    """
    Return True if the path contains the partition(s) provided.

    :param file_path:
    :param partitions:
    :return:
    """
    # Convert from a list of string to a dictionary of partition, operator &
    # value
    query_partitions = [
        potential.as_dict() for potential in
        [
            SplitString(x)
            for x in partitions
        ] if potential.valid
    ]

    return _meets_partition_params(file_path, query_partitions)


def _meets_partition_params(
    file_path: str,
    query_partitions: Optional[List[Dict[str, str]]],
):
    """
    Return True if the path contains the partition(s) provided.

    :param file_path:
    :param query_partitions:
    :return: True if meets the criteria.
    """

    # Copied from Consumption

    if query_partitions is None:
        return True

    # Sanitse the user input.
    for x in query_partitions:
        x['partition'] = x['partition'].strip()

    # Convert Lists of List to Dictionary using dictionary comprehension.
    found_partitions = {
        x[0]: x[1] for x in _get_partitions_in_filepath(file_path)
    }

    filtered = [
        x for x in query_partitions
        if x['partition'] in found_partitions
    ]

    # Example:
    # found_p = dict[(k,v), (k1,v1)] = k:v, k1:v1
    # query_p = [{'partition':'date', 'operator':'<', 'value':'20190102'}]

    for filterable in filtered:
        field = filterable['partition'].strip()
        compare_val = found_partitions[field]
        op = filterable['operator'].strip()
        filter_value = filterable['value'].strip()

        if field.startswith('as_of_'):
            try:
                filter_value_date = isoparse(filter_value)
            except ValueError as e:
                # This means a user has given an invalid date i.e. not in
                # the correct ISO 8601 format for Python datetime.
                raise ValueError(
                    'Was unable to parse the filter value you provided for '
                    'the `as_of_date` into a Python datetime: '
                    f"'{filter_value}', file_path: {file_path}'"
                ) from e

            try:
                compare_val = isoparse(compare_val)
            except ValueError:
                # Can not meet partition params as it not a date
                trace_logger.warning(
                    f'{file_path} is not a valid date.',
                    extra={
                        'file_path': file_path,
                        'filterable': filterable
                    },
                    exc_info=True
                )

                return False

            filter_value = filter_value_date

        match = eval_logical(
            op,
            field=compare_val,
            val=filter_value
        )

        if not match:
            # Short circuit as soon as we fail to match a filter.
            trace_logger.debug(
                f"Excluding file with path '{file_path}' because it "
                f"contains the partitions '{found_partitions}' and the "
                f"user is filtering with '{field}{op}{filter_value}' "
                f"which for this path is {compare_val}'."
            )
            return False

    not_filtered = [
        x for x in query_partitions
        if not x['partition'] in found_partitions
    ]

    if not_filtered:
        # trace_logger.warning(
        #     f"These query partitions '{not_filtered}' were not found as "
        #     f"keys in the S3 path '{file_path}', so we are going to "
        #     'let this file through the filter but either the user has '
        #     'supplied an partition that does not exist or one of the S3 '
        #     'paths does not follow the partition pattern of the first S3 '
        #     'path in this instance.'
        # )
        return False

    return True


def get_most_recent_common_prefix(
    common_prefixes: Iterator[str],
) -> List[str]:
    """
    This filtering logic only applies to Structured datasets. For a
    Structured dataset, we expect the first common prefix on S3 to be an
    `as_of_` representing when the data was published. It is expected to be
    in ISO 8601 format as this can be parsed into a Python DateTime.

    Return only the most recent `as_of` common prefix.

    :param common_prefixes: List of common prefixes for this dataset on S3.
    :return:  The most recent common_prefix for this dataset on S3. We return
        a collection because there is an edge case where there are common
        prefixes on S3 that do not have as_of, in which case
    """

    as_of_to_common_prefix: dict = defaultdict(list)
    as_of_warning_shown = False

    # Accumulate the `as_of_` dates from the common prefixes.
    for common_prefix in common_prefixes:
        partitions = {
            k.lower(): v for k, v in dict(
                _get_partitions_in_filepath(common_prefix)
            ).items()
        }

        # Partition cannot be filtered out if does not have an
        # `as_of_`. Get the first instance of a key that begins
        # with `as_of_`.
        as_of_key = next(
            (k for k, v in partitions.items()
             if k.startswith('as_of_')),
            None
        )

        if not as_of_key:
            # There is not an `as_of_` in this common prefix!
            # Only show the message the first time, otherwise it will
            # spam the logs.
            if not as_of_warning_shown:
                trace_logger.warning(
                    'as_of_ not in partitions but load type is full, '
                    'so not filtering out this common prefix but will '
                    'consider it as having the oldest as_of_ possible.',
                    extra={
                        'common_prefix': common_prefix,
                    }
                )
                as_of_warning_shown = True

            as_of_to_common_prefix[
                datetime.min
            ].append(common_prefix)
        else:
            # For a `load_type` = `Full Load` dataset, we only want
            # to return the data in the most recent `as_of_`
            # directory. We accumulate a dict of `as_of_` to
            # common_prefix, then at the end we know what the most recent
            # `as_of_` is and only return that common prefix.
            as_of_to_common_prefix[
                isoparse(partitions[as_of_key])
            ].append(common_prefix)

    if as_of_to_common_prefix:
        # Return only the most recent common prefix.
        # To sort, we have to action the iterator and therefore the type
        # changes to a List.
        most_recent_as_of = sorted(
            as_of_to_common_prefix.keys()
        )[-1]
        trace_logger.debug(
            'Load type is full, so returning most recent '
            f"as_of_ common prefix: '{most_recent_as_of}'"
        )
        return as_of_to_common_prefix[most_recent_as_of]
    else:
        # Edge case - the `load_type` is `Full Load` but
        # there are zero as_of_ keys, so return and empty list. The
        # code calling this filter will have to return a user
        # friendly error message about there being no files in the
        # dataset.
        # Alternatively, the cache might not be up to date causing
        # the most recent as_of_ to not be in the list.
        trace_logger.warning(
            'Zero as_of_ keys in the partitions but load type is '
            'full. We may have returned some paths that did not '
            'have an as_of_ key in their common_prefix. Alternatively, '
            'we may need to wait for the cache to update.'
        )
        return []


def get_common_prefixes(
    resource,
    bucket,
    prefix: str,
) -> Iterator[str]:
    """
    Abstract over the S3 location and return a list of the common prefixes
    (the top-level of directories under the dataset's S3 prefix..

    :param prefix: e.g. as_of_date=2019
    :param bucket: e.g. consumption-performance-test
    :param resource:

    :return: Iterator of paths for the top-level of directories under the
        dataset's S3 prefix.
    """
    trace_logger.debug(
        'get_common_prefixes - start paginating s3 list from prefix '
        f"'{prefix}' to delimiter '/'",
        extra={
            'Bucket': bucket.name,
            'Prefix': prefix,
        }
    )

    delimiter = '/'
    start_after = None  # Used for pagination of S3 results.
    max_loops = 1000000  # Used to prevent infinite loops.
    i = 0

    while i < max_loops and True:
        i += 1  # Counter to prevent infinite loops.
        response: Dict[str, str]

        if start_after:
            trace_logger.debug(
                f'next page of s3 list results (start_after {start_after})...',
                extra={
                    'Bucket': bucket.name,
                    'Prefix': prefix,
                    'Delimiter': delimiter,
                    'start_after': start_after,
                }
            )
            response = resource.meta.client.list_objects_v2(
                Bucket=bucket.name,
                Prefix=prefix,
                Delimiter=delimiter,
                StartAfter=start_after,
            )
        else:
            # Initial loop
            trace_logger.debug(
                'start paginating s3 list...',
                extra={
                    'Bucket': bucket.name,
                    'Prefix': prefix,
                    'Delimiter': delimiter,
                }
            )
            response = resource.meta.client.list_objects_v2(
                Bucket=bucket.name,
                Prefix=prefix,
                Delimiter=delimiter,
                # No StartAfter param in initial loop.,
            )

        # Type could be horrible as it comes from a response.
        common_prefixes: Union[
            str,
            Dict[str, str],
            List[Dict[str, str]]
        ] = response.get('CommonPrefixes', [])

        last_prefix = None
        if common_prefixes:
            if isinstance(common_prefixes, str):
                trace_logger.warning(
                    'common prefix came back as a single string '
                    'instead of a list of dict which was unexpected '
                    'but was handled'
                )
                last_prefix = common_prefixes
                yield common_prefixes
            if isinstance(common_prefixes, Dict):
                trace_logger.warning(
                    'common prefix came back as a single dict '
                    'instead of a list of dict which was unexpected '
                    'but was handled',
                    extra={
                        'response first N chars': str(response)[:1000]
                    }
                )
                last_prefix = common_prefixes.get(
                    'Prefix',
                    None
                )
                yield common_prefixes.get('Prefix', '')
            elif isinstance(common_prefixes, List):
                # Example of common_prefixes structure:
                #  [
                #     {
                #         "Prefix":
                #             "short_code/as_of_date=2020-09-22/"
                #     },
                #     {
                #         "Prefix":
                #             "short_code/as_of_date=2020-09-23/"
                #     }
                # ]
                for current_common_prefix in common_prefixes:
                    last_prefix = current_common_prefix.get(
                        'Prefix',
                        None
                    )
                    yield last_prefix
            else:
                trace_logger.error(
                    'type error for common_prefixes',
                    extra={
                        'common_prefixes': common_prefixes,
                        'type common_prefixes': type(common_prefixes),
                        'response first N chars': str(response)[:1000]
                    }
                )
                raise Exception(
                    'type error for common_prefixes - '
                    f'type: {type(common_prefixes)}, '
                    f'common_prefixes: {common_prefixes}'
                )
        else:
            trace_logger.debug(
                'Zero common_prefixes',
                extra={
                    'response first N chars': str(response)[:1000],
                }
            )
            last_prefix = None

        if response.get('IsTruncated') and last_prefix:
            start_after = last_prefix
            trace_logger.debug(
                'Truncated response, so more results on the next '
                'page',
                extra={
                    'start_after': start_after,
                    'response first N chars': str(response)[:1000],
                }
            )
        else:
            break


def is_hidden_file(path: str, content_type: str) -> bool:
    """
    [DL-4545][DL-4536][DL-5209] Do not read into files or
    directories that are cruft from Spark which Spark will ignore
    on read, e.g. files/dirs starting with `.` or `_` are hidden
    to Spark.

    Skip as_of_*=latest as it is a Spark temporary folder.

    :param content_type: Structured/Unstructured
    :param path: e.g. as_of_date=2019, .latest, as_of_date=latest
    :return: True if the path matches the criteria of a hidden file.
    """
    return (
        path.startswith('.') or
        path.startswith('_') or
        (
            path == 'metadata' and
            content_type != "Unstructured"
        ) or
        (
            path.startswith('as_of_') and
            path.endswith('=latest')
        )
    )


def __to_s3_proxy_path_and_size(
    object_summary,
    absolute_path: bool,
    organisation_short_code: str,
) -> Tuple[str, int]:
    if absolute_path:
        path = f"s3://{organisation_short_code}/{object_summary.key}"
        return path, object_summary.size
    else:
        return object_summary.key, object_summary.size


def get_s3_list(
    bucket,
    prefix: str,
    absolute_path: bool,
    organisation_short_code: str,
) -> Iterator[Tuple[str, int]]:
    """
    S3 list everything in this bucket under this prefix.

    :param bucket:
    :param prefix: e.g. abc/as_of_date=2019, abc/as_of_date
    :param absolute_path: True returns absolute path to the file on S3 proxy.
    :param organisation_short_code:
    :return: S3 proxy path and size.
    """

    trace_logger.debug(f"Get S3 list for prefix '{prefix}'.")

    object_summaries = bucket.objects.filter(
        # Prefix searches for exact matches and folders
        Prefix=prefix
    )

    # Convert each object_summary into an S3 proxy path.
    for object_summary in object_summaries:
        if not object_summary.key.endswith('/'):
            yield __to_s3_proxy_path_and_size(
                object_summary=object_summary,
                absolute_path=absolute_path,
                organisation_short_code=organisation_short_code
            )


def filter_non_hidden_common_prefixes(
    common_prefixes: Iterator[str],
    prefix: str,
    content_type: str,
) -> Iterator[str]:
    """
    Filter out S3 paths that are hidden directories. This function is intended
    for use with common prefixes.

    :param common_prefixes: Iterator of paths for the top-level of
        directories under the dataset's S3 prefix.
    :param prefix: e.g. abc/as_of_date=2019, abc/as_of_date
    :param content_type: Structured/Unstructured
    :return:
    """

    def shorten(common_prefix: str):
        """
        Path is made by stripping the common prefix of the prefix
        and the delimiter, which S3 returns.

        :param common_prefix: e.g. abc/as_of_date=2019/
        :return: as_of_date=2019
        """

        top_level_path = common_prefix
        if common_prefix.startswith(prefix):
            top_level_path = common_prefix[len(prefix):]
        if common_prefix.endswith('/'):
            top_level_path = top_level_path[:-1]
        return top_level_path

    yield from (
        common_prefix for common_prefix in common_prefixes
        if not is_hidden_file(shorten(common_prefix), content_type)
    )


def get_s3_list_filter_out_hidden(
    resource,
    bucket,
    prefix: str,
    content_type: str,
    load_type: str,
    absolute_path: bool,
    organisation_short_code: str,
) -> Iterator[Tuple[str, int]]:
    """
    Top-level common prefix filtering. We need to prevent reading into
    hidden folders. In the case of Bilateral the latest folders are
    growing linearly and for us to do an iterative list and post filter
    out folders we do not want is taking too long on prod. We need to
    filter out at the common prefix level!

    :param resource:
    :param bucket:
    :param prefix: e.g. abc/as_of_date=2019, abc/as_of_date
    :param content_type: Structured/Unstructured
    :param load_type: Either None, "Incremental Load" or "Full Load". None
        should have the same behaviour as "Incremental Load".
    :param absolute_path: True returns absolute path to the file on S3 proxy.
    :param organisation_short_code:
    :return: S3 proxy path and size.
    """

    common_prefixes: Iterator[str] = get_common_prefixes(
        resource=resource,
        bucket=bucket,
        prefix=prefix,
    )

    non_hidden_common_prefixes = filter_non_hidden_common_prefixes(
        common_prefixes=common_prefixes,
        prefix=prefix,
        content_type=content_type,
    )

    unique_non_hidden_common_prefixes = set([])

    trace_logger.debug(
        f"load_type from Catalogue: '{load_type}'",
        extra={
            'load_type': load_type,
        }
    )

    if load_type == 'Full Load':
        # For a Structured dataset, we expect the first common prefix on S3
        # to be an `as_of` representing when the data was published.
        #
        # If the dataset's load type is "Full Load"
        # then return only data from the most recent `as_of` common prefix.
        # Else (if the dataset's load type is "Incremental Load" or None),
        # then we return data from all as_of common prefixes.
        unique_non_hidden_common_prefixes.update(get_most_recent_common_prefix(
            common_prefixes=non_hidden_common_prefixes
        ))
    else:
        for common_prefix in non_hidden_common_prefixes:
            # We do not need the entire path of each non-hidden
            # common prefix. We only need to know the prefix and the
            # partition name of the common prefix (if exists) knowing
            # it will not be a `.` or `_`. We do not want to keep the
            # unique value of the partition as that will cause use to
            # make an S3 call for each unique partition.
            # e.g. for common_prefix
            #     abc/as_of_date=2020
            # we only store
            #     abc/as_of_date
            # and so all the as_of_dates are collected in a single
            # S3 list operation with prefix starting abc/as_of_date.
            if '=' in common_prefix:
                prefix_plus_partition, _ = common_prefix.split('=', 1)
            else:
                # Edge case where the top-level is not partitioned.
                prefix_plus_partition = common_prefix

            unique_non_hidden_common_prefixes.add(prefix_plus_partition)

    for non_hidden_common_prefix in unique_non_hidden_common_prefixes:
        # Now loop over the unique prefix plus partition to get a list
        # of the contents.
        # e.g. for
        #   abc/as_of_date
        # the S3 list could return:
        #   abc/as_of_date=2019/file.parquet
        #   abc/as_of_date=2020/file.parquet
        yield from get_s3_list(
            bucket=bucket,
            prefix=non_hidden_common_prefix,
            absolute_path=absolute_path,
            organisation_short_code=organisation_short_code,
        )



import nox
import os


def test_existing_requirements(session):
    session.install('-r', '../../requirements.txt')
    session.install('-r', '../../requirements-test.txt')
    os.chdir('../../')
    session.run('pytest', '-m', 'not integration', '--verbose')


def test_requirements_sc0(session):
    """
        Smoke test - single unittest, no integration tests.
        Integration tests are run in test_requirements_offset_scenarios.
    """
    TEST_FILES = ['./tests/models/test_structured_dataset.py']
    session.install('-r', '../../requirements.txt')
    session.install('-r', '../../requirements-test.txt')
    os.chdir('../../')
    session.run('pytest', '-m', "'not integration'", '--verbose', *TEST_FILES)


def test_requirements_offset_scenarios(session):
    """
        Run all unit tests and integration tests
        using modified requirements.txt, where for each
        package>=version_number, version_number is found given its
        proximity to MIN_DATE environment variable.
    """
    scenario = os.environ.get('NOX_SCENARIO')
    min_date = os.environ.get('MIN_DATE')
    install_betas: bool = os.environ.get('INSTALL_BETAS', None) == 'True'

    if scenario == '1':
        print('Not using a min_date')
        min_date = None
    from_file = '../../requirements.txt'
    to_file = f'requirements-{scenario}.txt'

    rm = RequirementsModifier(from_file)
    rm.modify_requirements(from_file, to_file, min_date)
    if install_betas:
        print('Install beta versions of dependencies')
        session.install('-r', f'requirements-{scenario}.txt', '--pre')
        session.install('-r', '../../requirements-test.txt', '--pre')
    else:
        print('Install release versions of dependencies')
        # TODO scott: a bit of tech debt - we didn't need to write to a file,
        #  doing a session.install you can install a dependency just like
        #  you would with pip. So instead of adding a line to a file, just
        #  session install and print what version and date you are using and
        #  the command line will print the result of the install anyway).
        session.install('-r', f'requirements-{scenario}.txt')
        session.install('-r', '../../requirements-test.txt')

    session.run('pip', 'list', '--format', 'columns')
    os.chdir('../../')
    session.run('pytest', '-m', 'not integration', '--verbose')


PYTHON_VERSIONS_LIST = [
    '3.6.10',
    '3.7',
    '3.8.1',
    # '3.9.1', # Not yet supported by Pyarrow
]
PIP_VERSIONS_LIST = [
    '19.0',  # Oldest pip working with pyarrow 3.0.0. 22nd January 2019.
    '21.0.1',  # Latest pip on 30th January 2021.
]


def test(session, pip_version):
    session.run('pip', 'install', '-U', 'pip=={}'.format(pip_version))

    scenarios_dict = {
        '0': test_requirements_sc0,
        '1': test_requirements_offset_scenarios,
        '2': test_requirements_offset_scenarios,
    }
    scenario = os.environ.get('NOX_SCENARIO')
    print('NOX_SCENARIO=', scenario)
    if scenario in scenarios_dict:
        scenarios_dict[scenario](session)
    else:
        test_existing_requirements(session)


@nox.session(python=PYTHON_VERSIONS_LIST)
@nox.parametrize('pip_version', PIP_VERSIONS_LIST)
def test_python36_python37(session, pip_version):
    min_date = os.environ.get('MIN_DATE')
    if session.python == '3.8.1' and min_date <= '2020-02-01':
        print(
            'Skip as Python 3.8.1 was not supported by our dependencies before'
            f'this date. Python {session.python}, min_date {min_date}'
        )
        session.skip()

    test(session, pip_version)


class RequirementsModifier:

    def __init__(self, from_file):
        self.requirements = from_file
        self.altered_reqs = []

        # Following set for each individual package.
        self.package = None
        self.current_version = None
        self.date_to_version = None
        self.version_to_date = None
        self.dates_lst = None
        self.version_lst = None
        self.min_date = None
        self.passed_constraints = False

        # Temporary dicts to seek out potential versioning issues
        self.version_limits = None
        self.version_override = None
        self.version_omit = None

    def set_known_issues(self):
        # these constraints should be added to requirements.txt
        # if they help avoid versioning error

        # These are the earliest versions of the dependencies that we can
        # support e.g. we cannot support numpy below `1.16.0rc1` because
        # pandas would not be compatible.
        # package<version
        self.version_limits = {
            'numpy': '1.16.0rc1',  # 2018-12-20
            'Flask': '1.0.2',  # 2018-05-02
            'pyarrow': '0.11.1',  # 2018-10-24
            'keyring': '17.1.1',  # 2018-12-23
            'pypermedia': '0.4.2',  # 2016-02-18
            'python-json-logger': '0.1.10',  # 2018-11-07
            'requests': '2.22.0',  # 2019-05-16
            'deprecated': '1.2.4',  # 2019-05-16
            'tqdm': '4.28.1',  # 2018-10-21
            'pyjwt': '1.7.1',  # 2018-12-07
            'requests-toolbelt': '0.8.0',  # 2017-05-20
            'pandas': '0.25.3',  # 2019-11-01
            'tabulate': '0.8.2',  # 2017-11-26
            'pyhumps': '0.5.1',  # 2018-10-30
            'wrapt': '1.10.11',  # 2017-08-11
            'ipywidgets': '7.4.2',  # 2018-09-14
            'ipython': '7.2.0',  # 2018-11-30
            'python-dateutil': '2.7.5',  # 2018-10-27
            's3transfer': '0.1.13',  # 2018-02-15
            # IPV6_ADDRZ_RE should be available in urllib3 1.25.4 forward
            # which is the lowest version of the library botocore supports.
            # https://github.com/boto/botocore/issues/2186#issuecomment-712449171
            'urllib3': '1.25.4',  # 2019-05-02
            'botocore': '1.12.156',  # 2019-05-04 Allows urllib 1.25.4
            'boto3': '1.9.71',  # 2018-12-21
            'boto': '2.49.0',  # 2018-07-11
        }

        # package==version
        self.version_override = {}

        # Ignore these versions as they have breakages e.g. numpy `1.20.0rc1`
        # is not compatible with pyarrow/pandas released at the same time.
        # package!=version
        self.version_omit = {
            'numpy': '1.20.0rc2',
            'pyarrow': '0.13.0',
        }

    def set_requirement_details(self):
        import requests
        dc = requests.get(f'https://pypi.org/pypi/{self.package}/{self.current_version}/json').json()
        releases = dc['releases']
        date_to_version, dates_lst = self.get_requirement_date_to_version(
            releases)
        version_to_date, version_lst = self.get_requirement_version_to_date(
            releases)

        self.date_to_version = date_to_version
        self.version_to_date = version_to_date
        self.dates_lst = dates_lst
        self.version_lst = version_lst

    def get_requirements_list(self, file):
        with open(file) as f:
            ls = f.readlines()
            ln = []
            for l in ls:
                sl = l.split(' ')
                if len(sl) == 1:
                    sl = l.split('\n')
                ln.append(sl[0])
        return ln

    def get_requirement_date_to_version(self, releases):
        dates_lst = []
        dates_dct = {}
        for k, v in releases.items():
            try:
                dates_dct[v[0]['upload_time']] = k
                dates_lst.append(v[0]['upload_time'])
            except Exception:
                try:
                    dates_dct[k] = v['upload_time']
                    dates_lst.append(v[0]['upload_time'])
                except Exception:
                    pass
                    # print(f'No release information for {k}, date to version')
        dates_lst.sort(reverse=True)
        return dates_dct, dates_lst

    def get_requirement_version_to_date(self, releases):
        versions_lst = []
        versions_dict = {}
        for k, v in releases.items():
            try:
                versions_dict[k] = v[0]['upload_time']
                versions_lst.append(k)
            except Exception:
                try:
                    versions_dict[v['upload_time']] = k
                    versions_lst.append(k)
                except Exception:
                    pass
                    # print(f'No release information for {k}, version to date')
        versions_lst.sort(reverse=True)
        return versions_dict, versions_lst

    def write_requirements(self, reqs, file):
        with open(file, 'w') as f:
            f.writelines(reqs)

    def print_requirements_diff(self):
        print('requirements diff:')
        rlen = len(self.altered_reqs)
        r = 0
        while r < rlen:
            print(f'{self.requirements[r]} -> {self.altered_reqs[r]}')
            r += 1

    def get_closest_earlier_date(self, min_date):
        for date in self.dates_lst:
            if date < min_date:
                upd_vrsn = self.date_to_version[date]
                return upd_vrsn, date

        last_date = self.dates_lst[-1:][0]
        return self.date_to_version[last_date], last_date

    def get_earlier_version_with_offset(self, vrsn, offset):
        # 1. get date for existing version
        try:
            date = self.version_to_date[vrsn]
        except Exception:
            for v in self.version_lst:
                if v.startswith(vrsn):
                    # print(f'Using {v} instead of {vrsn} for package: {pckg}')
                    vrsn = v
                    date = self.version_to_date[vrsn]

        year = int(date[:4]) - int(offset[:4])
        month = int(date[5:7]) - int(offset[5:7])
        if month < 10:
            month = f'0{month}'
        day = int(date[8:10]) - int(offset[8:10])
        if day < 10:
            day = f'0{day}'

        udate = f'{year}-{month}-{day}'

        self.min_date = os.environ.get('MIN_DATE')
        if self.min_date is not None and udate < self.min_date:
            udate = self.min_date

        return self.get_closest_earlier_date(udate)

    def find_acceptable_version(self, constraints_dict):
        from packaging import version  # installed to base image

        def run_all_constraint_checks(alt_vrsn, alt_date):
            # Check constraints within requirements.txt
            if ('!=' in constraints_dict and
                    any(v == alt_vrsn for v in constraints_dict['!='])):
                # move down 1 version
                alt_vrsn, alt_date = not_equal(alt_vrsn)

            if ('<=' in constraints_dict and
                    version.parse(alt_vrsn) > version.parse(constraints_dict['<='][0])):
                alt_vrsn, alt_date = smaller_equal()

            if ('<' in constraints_dict and
                    version.parse(alt_vrsn) >= version.parse(constraints_dict['<'][0])):
                alt_vrsn, alt_date = smaller()

            # Apply temporary nox testing constraints
            # 1. check if req has a known issue / limit in version number for
            # nox tests
            if (self.package in self.version_limits and
                    version.parse(alt_vrsn) < version.parse(self.version_limits[self.package])):
                alt_vrsn = self.version_limits[self.package]
                alt_date = self.version_to_date[alt_vrsn]

            # 2. override modification to specific version
            if self.package in self.version_override:
                alt_vrsn = self.version_override[self.package]
                alt_date = self.version_to_date[alt_vrsn]
                print(
                    'override modification to specific version. '
                    f'alt_vrsn {alt_vrsn}, alt_date {alt_date}'
                )

            # 3. for pckg!=version find an earlier version number
            if (self.package in self.version_omit and
                    alt_vrsn == self.version_omit[self.package]):
                alt_vrsn, alt_date = self.get_closest_earlier_date(alt_vrsn)

            self.passed_constraints = True
            return alt_vrsn, alt_date

        def skip_modification():
            # skip modification and check existing version in requirements.txt
            # is valid for ==
            # not all ~=versionA would work with ==versionA!
            try:
                alt_date = self.version_to_date[self.current_version]
                alt_vrsn = self.current_version
                self.passed_constraints = True
                return alt_vrsn, alt_date

            except:
                # req with ~= might be missing a subversion
                # number so we find correct version
                for v in self.version_lst:
                    if self.current_version in v:  # 1.2 in 1.2.9
                        alt_date = self.version_to_date[v]
                        alt_vrsn = v
                        return alt_vrsn, alt_date

        def smaller_equal():  # <=
            # altered req is too small, boundary condition,
            alt_vrsn = constraints_dict['<='][0]
            self.passed_constraints = False
            return alt_vrsn, alt_date

        def smaller():  # <
            # altered req is too small, boundary condition,
            # this occurs when the existing req is 1 off from being too small
            alt_vrsn, alt_date = skip_modification()
            self.passed_constraints = False
            return alt_vrsn, alt_date

        def not_equal(alt_vrsn):  # !=
            # skip the erroneous version
            alt_vrsn, alt_date = self.get_closest_earlier_date(self.version_to_date[alt_vrsn])
            self.passed_constraints = False
            return alt_vrsn, alt_date

        # Get valid version
        if self.min_date is None:
            alt_vrsn, alt_date = skip_modification()

        else:
            # Find a valid earlier date than the environment variable min_date
            alt_vrsn, alt_date = self.get_closest_earlier_date(self.min_date)

            while self.passed_constraints is False:
                alt_vrsn_, alt_date = run_all_constraint_checks(alt_vrsn, alt_date)

                # a new alt_vrsn must be tested again with the constraints
                if alt_vrsn_ != alt_vrsn:
                    self.passed_constraints = False
                alt_vrsn = alt_vrsn_

        return alt_vrsn, alt_date

    def modify_requirements(self, from_file, to_file, min_date):
        # init overrides and set lower limits for package versions to avoid
        # known issues and incompatibilities.
        self.set_known_issues()
        self.min_date = min_date
        # fetch original requirements to modify.
        self.requirements = self.get_requirements_list(from_file)

        for req in self.requirements:
            # I. process a line of the requirements.txt file
            cmpsymbols = ['<=', '>=', '==', '!=', '~=', '*', '<', '>']

            def get_cmpsym(sub):
                for c in cmpsymbols:
                    if c in sub:
                        return c

            L = req.split(' ')
            if L[0].startswith('#'):
                continue
            subexpr = L[0].split(',')

            # subexpr[0] for nox modifications (>=, ==, ~=)
            se = subexpr.pop(0)
            cmpsym = get_cmpsym(se)
            E = se.split(cmpsym)
            self.package = E[0]
            self.current_version = E[-1]

            # request all versions and release dates of a given requirement as helper variables.
            self.set_requirement_details()

            # subexpr[1].. constraints for erroneous versions (<=, !=)
            cntr_dict = {}
            for se in subexpr:
                c = get_cmpsym(se)
                E = se.split(c)
                if c not in cntr_dict:
                    cntr_dict[c] = []
                cntr_dict[c].append(E[-1])

            # II. Do modification to package version.
            self.passed_constraints = False
            alt_vrsn, alt_date = self.find_acceptable_version(cntr_dict)

            if self.min_date is None:
                if self.package in self.version_omit:
                    # Edge case where we are testing betas and we find that
                    # a beta causes a problem so we neeed to exclude it.
                    self.altered_reqs.append(f'{self.package}>={alt_vrsn},!={self.version_omit[self.package]}    # {alt_date[:10]}\n')
                else:
                    self.altered_reqs.append(f'{self.package}>={alt_vrsn}    # {alt_date[:10]}\n')
            else:
                self.altered_reqs.append(f'{self.package}=={alt_vrsn}    # {alt_date[:10]}\n')

        # III. write modified requirements to file for nox testing
        self.write_requirements(self.altered_reqs, to_file)
        self.print_requirements_diff()
